from typing import NamedTuple
from jaxtyping import Float, Complex, Array

class LinalgDecomposition(NamedTuple):
    values: Complex[Array, "values"]
    vectors: Complex[Array, "n vectors"]

class RealLinalgDecomposition(NamedTuple):
    values: Float[Array, "values"]
    vectors: Float[Array, "n vectors"]


from sklearn.utils.estimator_checks import parametrize_with_checks
from sys import path
path.append('../')
from kooplearn.kernels import Kernel
from kooplearn.estimators import ReducedRank, PrincipalComponent

import numpy as np

class PaddedLinearKernel(Kernel):
    """Dummy kernel for testing. If X and Y have the same number of features is a linear kernel. Otherwise the smaller of the two is padded with zeros and the other is linearly combined with the padded one."""
    def __init__(self):
        pass
    def __call__(self, X, Y=None, backend='numpy'):
        if Y is None:
            lin_X = np.dot(X, X.T)
            return lin_X
        if Y is not None:
            if Y.shape[1] == X.shape[1]:
                pass
            elif X.shape[1] < Y.shape[1]:
                _zeroes = np.zeros((X.shape[0], Y.shape[1] - X.shape[1]))
                X = np.c_[X, _zeroes]  
            else:
                _zeroes = np.zeros((X.shape[0], X.shape[1] - Y.shape[1]))
                Y = np.c_[Y, _zeroes]
            lin_XY = np.dot(X, Y.T)
            return lin_XY
                
kernel = PaddedLinearKernel()
parameters = {
    'kernel': kernel,
    'svd_solver': 'full',
    'rank': 1
}
tikhonov_reg = None
@parametrize_with_checks([ReducedRank(**parameters, tikhonov_reg=tikhonov_reg), PrincipalComponent(**parameters)])
def test_sklearn_compatible_estimator(estimator, check):
    check(estimator)


from typing import Optional
import torch
from torch import Tensor
from einops import einsum

def generalized_eigh(A: Tensor, B: Tensor) -> tuple:
     #A workaround to solve a real symmetric GEP Av = \lambda Bv problem in JAX. (!! Not numerically efficient)
     Lambda, Q = torch.linalg.eigh(B)
     rsqrt_Lambda = torch.diag(Lambda.rsqrt())
     sqrt_B = Q@rsqrt_Lambda
     _A = 0.5*(sqrt_B.T@(A@sqrt_B) + sqrt_B.T@((A.T)@sqrt_B)) #Force Symmetrization
     values, _tmp_vecs = torch.linalg.eigh(_A) 
     vectors = Q@(rsqrt_Lambda@_tmp_vecs)
     return values, vectors

def spd_norm(vecs: Tensor, spd_matrix: Tensor) -> Tensor:
     _v = torch.mm(spd_matrix, vecs)
     _v_T = torch.mm(spd_matrix.T, vecs)
     return torch.sqrt(0.5*torch.linalg.vecdot(vecs, _v + _v_T, dim = 0).real)

def reduced_rank_regression(
    input_covariance: Tensor,
    cross_covariance: Tensor, #C_{XY}
    tikhonov_reg: float,
    ) -> tuple:
    
    n = input_covariance.shape[0]
    reg_input_covariance = input_covariance + tikhonov_reg*torch.eye(n, dtype=input_covariance.dtype, device=input_covariance.device)

    _crcov = torch.mm(cross_covariance, cross_covariance.T)
    _values, _vectors = generalized_eigh(_crcov, reg_input_covariance) 
    
    _norms = spd_norm(_vectors, reg_input_covariance)
    vectors = _vectors*(1/_norms)
    return _values, vectors

def tikhonov_regression(
    input_covariance: Tensor,
    tikhonov_reg: float,
    ) -> tuple:
    n = input_covariance.shape[0]
    reg_input_covariance = input_covariance + tikhonov_reg*torch.eye(n, dtype=input_covariance.dtype, device=input_covariance.device)

    Lambda, Q = torch.linalg.eigh(reg_input_covariance)
    return Lambda, Q@torch.diag(Lambda.rsqrt())

def eig(
    fitted_estimator: Tensor, 
    cross_covariance: Tensor,
    rank: Optional[int] = None,
    ) -> tuple:
      
    if rank is not None:
        _, idxs = torch.topk(fitted_estimator.values, rank)
        U = fitted_estimator[:, idxs]
    else:
        U = fitted_estimator
    
    #U@(U.T)@Tw = v w -> (U.T)@T@Uq = vq and w = Uq 
    values, Q = torch.linalg.eig((U.T)@(cross_covariance@U))
    return values, U@Q


def sq_error(
    input_data: Tensor,
    output_data: Tensor,
    estimator: Tensor
    ) -> float:
    """Mean squared error between the output data and the estimator applied to the input data.

    Args:
        input_data (Tensor): Tensor of shape (n, dim) containing the input data.
        output_data (Tensor): Tensor of shape (n, dim) containing the output data.
        estimator (Tensor): Tensor of shape (dim, dim) containing the estimator.

    Returns:
        float: mean squared error between the output data and the estimator applied to the input data.
    """    
    err = output_data.T - torch.mm(estimator.T, input_data.T)
    return torch.mean(torch.linalg.vector_norm(err, ord=2, dim=0)**2)

def naive_predict(
    featurized_x: Tensor,
    fitted_estimator: Tensor, 
    input_data: Tensor,
    output_raw_data: Tensor,
    rank: Optional[int] = None,
    ) -> Tensor:

    if rank is not None:
        _, idxs = torch.topk(fitted_estimator.values, rank)
        U = fitted_estimator[:, idxs]
    else:
        U = fitted_estimator

    x = einsum(
        featurized_x,
        U,
        U.T,
        input_data,
        output_raw_data,
        "d_in, d_in r, r d_out, n d_out, n dim -> dim"
    )
    num_data = float(input_data.shape[0])
    return (num_data**-1)*x


##Imports
from typing import Optional
import functools
#Scientific libs
import numpy as np
#JAX
import jax.scipy
import jax.scipy.optimize
import jax.numpy as jnp
import jax.lax
import jax

from jaxtyping import Float, Complex, Array

import optax
from torch.utils.tensorboard import SummaryWriter

#MISC
from einops import einsum 

from kooplearn.jax.typing import LinalgDecomposition, RealLinalgDecomposition
from kooplearn.jax.linalg import batch_spd_norm, generalized_eigh

def reduced_rank_regression(
    input_covariance: Float[Array, "d d"],
    cross_covariance: Float[Array, "d d"], #C_{XY}
    tikhonov_reg: float,
    ) -> RealLinalgDecomposition:
    """JAX-implementation of the reduced rank regression estimator.

    Args:
        input_covariance (Float[Array, &quot;d d&quot;]): Input Covariance matrix
        cross_covariance (Float[Array, &quot;d d&quot;]): Cross covariance matrix

    Returns:
        RealLinalgDecomposition: NamedTuple with the fitted estimator.
    """
    reg_input_covariance = input_covariance + tikhonov_reg*jnp.eye(input_covariance.shape[0], dtype=input_covariance.dtype)
    _crcov = cross_covariance@(cross_covariance.T) 
    _gep = generalized_eigh(_crcov, reg_input_covariance)
    
    _norms = batch_spd_norm(_gep.vectors, reg_input_covariance)
    vectors = _gep.vectors*(jax.lax.reciprocal(_norms))
    return RealLinalgDecomposition(_gep.values, vectors)

def randomized_reduced_rank_regression(
    input_covariance: Float[Array, "d d"],
    cross_covariance: Float[Array, "d d"], #C_{XY}
    tikhonov_reg: float,
    rank: int,
    key, #PRNGKey
    n_oversamples: int = 5, 
    iterated_power: int = 2,
    ) -> RealLinalgDecomposition:
    """JAX-implementation of the randomized reduced rank regression estimator.

    Args:
        input_covariance (Float[Array, &quot;d d&quot;]): Input Covariance Matrix
        cross_covariance (Float[Array, &quot;d d&quot;]): Cross covariance matrix
        rank (int): Rank of the estimator
        key (_type_): PRNG key
        iterated_power (int, optional): Number of power iterations to perform. Defaults to 2.

    Returns:
        RealLinalgDecomposition: NamedTuple with the fitted estimator.
    """    
    reg_input_covariance = input_covariance + tikhonov_reg*jnp.eye(input_covariance.shape[0], dtype=input_covariance.dtype)
    _crcov = cross_covariance@(cross_covariance.T)
    sketch = jax.random.normal(key, (reg_input_covariance.shape[0], rank + n_oversamples))

    for _ in range(iterated_power):
        _tmp_sketch = jax.scipy.linalg.solve(reg_input_covariance, sketch, assume_a='pos')
        sketch = _crcov@_tmp_sketch

    sketch_p =  jax.scipy.linalg.solve(reg_input_covariance, sketch, assume_a='pos')  
    F_0 = (sketch_p.T)@sketch
    F_1 = (sketch_p.T)@(_crcov@sketch_p)
    
    _gep = generalized_eigh(F_1, F_0)
    _norms = batch_spd_norm(_gep.vectors, F_0)
    vectors = _gep.vectors*(jax.lax.reciprocal(_norms))

    return RealLinalgDecomposition(_gep.values, sketch_p@vectors)

def tikhonov_regression(
    input_covariance: Float[Array, "d d"],
    tikhonov_reg: float,
    ) -> RealLinalgDecomposition:
    """JAX-implementation of the Tikhonov regression estimator.

    Args:
        input_covariance (Float[Array, &quot;d d&quot;]): Input Covariance matrix
        tikhonov_reg (float): Tikhonov regularization parameter

    Returns:
        RealLinalgDecomposition: NamedTuple with the fitted estimator.
    """    
    reg_input_covariance = input_covariance + tikhonov_reg*jnp.eye(input_covariance.shape[0], dtype=input_covariance.dtype)

    Lambda, Q = jnp.linalg.eigh(reg_input_covariance)
    rsqrt_Lambda = jnp.diag(jax.lax.rsqrt(Lambda))
    return RealLinalgDecomposition(Lambda, Q@rsqrt_Lambda)

def iterative_regression(
    input_data: Float[Array, "n d"],
    output_data: Float[Array, "n d"],
    num_iterations: int, 
    learning_rate: float=1.0, 
    momentum: bool = None, 
    nesterov: bool = False,
    tensorboard: bool = False
    ) -> Float[Array, "d d"]:    
    d = input_data.shape[1]
    estimator = jnp.zeros((d, d), dtype=input_data.dtype)

    optimizer = optax.sgd(learning_rate, momentum=momentum, nesterov=nesterov)
    opt_state = optimizer.init(estimator)

    @jax.jit
    def step(estimator, opt_state, input_batch, output_batch):
        loss_value, grads = jax.value_and_grad(sq_error, argnums = 2)(input_batch, output_batch, estimator)
        updates, opt_state = optimizer.update(grads, opt_state, estimator)
        estimator = optax.apply_updates(estimator, updates)
        return estimator, opt_state, loss_value
    if tensorboard:
        writer = SummaryWriter()
    for k in range(num_iterations):
        estimator, opt_state, loss_value = step(estimator, opt_state, input_data, output_data)
        if tensorboard and (k%10 == 0):
            writer.add_scalar("Loss/training", np.asarray(loss_value), k)
            _estim = np.asarray(estimator)
            _vals = np.linalg.eigvals(_estim)
            _leading_vals = np.sort_complex(_vals)[::-1][:3]
            writer.add_scalars("Eigenvalues/real_part", dict(zip(['1','2','3'], _leading_vals.real)), k)
            writer.add_scalars("Eigenvalues/imaginary_part", dict(zip(['1','2','3'], _leading_vals.imag)), k)
    return estimator

def eig(
    fitted_estimator: RealLinalgDecomposition, 
    cross_covariance: Float[Array, "d d"],
    rank: Optional[int] = None,
    ) -> LinalgDecomposition:
    """Eigenvalue decomposition of the fitted Koopman estimator.

    Args:
        fitted_estimator (RealLinalgDecomposition): NamedTuple with the fitted estimator.
        cross_covariance (Float[Array, &quot;d d&quot;]): Cross Covariance matrix.
        rank (Optional[int], optional): Rank of the estimator. Defaults to None meaning full rank. When rank is None, reduced_rank_regression and tikhonov_regression are equivalent and correspond to the EDMD estimator.

    Returns:
        LinalgDecomposition: Eigenvalue decomposition of the Koopman operator.
    """    
    if rank is not None:
        _, idxs = jax.lax.top_k(fitted_estimator.values, rank)
        U = (fitted_estimator.vectors)[:, idxs]
    else:
        U = fitted_estimator.vectors
    
    CPU_DEVICE = jax.devices('cpu')[0]
    #Moving arrays on the CPU as the nonsymmetric eig is not implemented on GPU
    U = jax.device_put(U, CPU_DEVICE)
    _xcov = jax.device_put(cross_covariance, CPU_DEVICE)
    #U@(U.T)@Tw = v w -> (U.T)@T@Uq = vq and w = Uq 
    values, Q = jnp.linalg.eig((U.T)@(_xcov@U))
    return LinalgDecomposition(values, U@Q)

@functools.partial(jax.vmap, in_axes=(0, 0, None), out_axes=0)
def sq_error_batch(
    input_features: Float[Array, "d"],
    output_features: Float[Array, "d"],
    estimator: Complex[Array, "d d"]
    ) -> float:
    err = output_features - jnp.dot(estimator.T, input_features)
    return jax.lax.square(jnp.linalg.norm(err))

def sq_error(
    input_data: Float[Array, "n d"],
    output_data: Float[Array, "n d"],
    estimator: Complex[Array, "d d"]
    ) -> float:
    """Mean squared error of the estimator. 

    Args:
        input_data (Float[Array, &quot;n d&quot;]): Initial data
        output_data (Float[Array, &quot;n d&quot;]): Evolved data
        estimator (Complex[Array, &quot;d d&quot;]): Estimator

    Returns:
        float: Mean squared error.
    """    
    return jnp.mean(sq_error_batch(input_data, output_data, estimator))

def naive_predict(
    featurized_x: Float[Array, "d"],
    fitted_estimator: RealLinalgDecomposition, 
    input_data: Float[Array, "n d"],
    output_raw_data: Float[Array, "n dim"],
    rank: Optional[int] = None,
    ) -> Float[Array, "dim"]:

    if rank is not None:
        _, idxs = jax.lax.top_k(fitted_estimator.values, rank)
        U = (fitted_estimator.vectors)[:, idxs]
    else:
        U = fitted_estimator.vectors
    x = einsum(
        featurized_x,
        U,
        U.T,
        input_data,
        output_raw_data,
        "d_in, d_in r, r d_out, n d_out, n dim -> dim"
    )
    num_data = float(input_data.shape[0])
    return jax.lax.reciprocal(num_data)*x


import numpy as np

from scipy.sparse.linalg import aslinearoperator, LinearOperator

from sklearn.utils import check_array, check_random_state

from warnings import warn

def sort_and_crop(vec, num_components = None):
    """Return the indices of the largest ``num_components`` elements in vec in descending order.

    Args:
        vec (ndarray): 1D array of floats
        num_component (int, optional): Number of indices to retain. Defaults to None corresponding to every indices.

    Returns:
        ndarray: array of integers corresponding to the indices of the largest num_components elements in vec.
    """
    assert np.ndim(vec) == 1, "'vec' must be a 1D array"
    sort_perm = np.argsort(vec)[::-1] # descending order
    if num_components is None:
        return sort_perm
    else:
        return sort_perm[:num_components]
    
def weighted_norm(A, M = None):
    r"""Weighted norm of the columns of A.

    Args:
        A (ndarray): 1D or 2D array. If 2D, the columns are treated as vectors.
        M (ndarray or LinearOperator, optional): Weigthing matrix. the norm of the vector :math:`a` is given by :math:`\langle a, Ma\rangle`. Defaults to None, corresponding to the Identity matrix. Warning: no checks are performed on M being a PSD operator. 

    Returns:
        (ndarray or float): If ``A.ndim == 2`` returns 1D array of floats corresponding to the norms of the columns of A. Else return a float.
    """    
    assert A.ndim <= 2, "'A' must be a vector or a 2D array"
    _1D = (A.ndim == 1)
    if M is None:
        norm = np.linalg.norm(A, axis=0)
    else:
        if _1D:
            _A = aslinearoperator(M).matvec(np.ascontiguousarray(A))
        else:
            _A = aslinearoperator(M).matmat(np.asfortranarray(A))  
        norm = np.real(
                np.sum(
                    np.conj(A)*_A, 
                    axis=0)
                )        
    return np.sqrt(norm)

def weighted_dot_product(A, B, M=None):
    """Weighted dot product between the columns of A and B. The output will be equivalent to :math:`A^{*} M B` if A and B are 2D arrays.

    Args:
        A, B (ndarray): 1D or 2D arrays.
        M (ndarray or LinearOperator, optional): Weigthing matrix. Defaults to None, corresponding to the Identity matrix. Warning: no checks are performed on M being a PSD operator. 

    Returns:
        (ndarray or float): The result of :math:`A^{*} M B`.
    """    
    assert A.ndim <= 2, "'A' must be a vector or a 2D array"
    assert B.ndim <= 2, "'B' must be a vector or a 2D array"
    A_adj = np.conj(A.T)
    _B_1D = (B.ndim == 1)
    if M is None:
        return np.dot(A_adj, B)
    else:
        if _B_1D:
            _B = aslinearoperator(M).matvec(np.ascontiguousarray(B))
        else:
            _B = aslinearoperator(M).matmat(np.asfortranarray(B))
        return np.dot(A_adj, _B)

def _column_pivot(Q, R, k, squared_norms, columns_permutation):
    """
        Helper function to perform column pivoting on the QR decomposition at the k iteration. No checks are performed. For internal use only.
    """
    _arg_max = np.argmax(squared_norms[k:])
    j = k + _arg_max
    _in = [k, j]
    _swap = [j,k]
    #Column pivoting
    columns_permutation[_in] = columns_permutation[_swap]
    Q[:, _in] = Q[:,_swap]
    R[:k, _in] = R[:k,_swap]
    squared_norms[_in] = squared_norms[_swap]
    return Q, R, squared_norms, columns_permutation

def modified_QR(A, M = None, column_pivoting = False, rtol = 2.2e-16, verbose = False):
    """Modified QR algorithm with column pivoting. Implementation follows the algorithm described in [1].

    Args:
        A (ndarray): 2D array whose columns are vectors to be orthogonalized.
        M (ndarray or LinearOperator, optional): PSD linear operator. If not None, the vectors are orthonormalized with respect to the scalar product induced by M. Defaults to None corresponding to Identity matrix.
        column_pivoting (bool, optional): Whether column pivoting is performed. Defaults to False.
        rtol (float, optional): relative tolerance in determining the numerical rank of A. Defaults to 2.2e-16. This parameter is used only when ``column_pivoting == True``.
        verbose (bool, optional): Whether to print informations and warnings about the progress of the algorithm. Defaults to False.

    Returns:
        tuple: A tuple of the form (Q, R), where Q and R satisfy A = QR. If ``column_pivoting == True``, the permutation of the columns of A is returned as well.
    
    [1] A. Dax: 'A modified Gram–Schmidt algorithm with iterative orthogonalization and column pivoting', https://doi.org/10.1016/S0024-3795(00)00022-7. 
    """    
    A = check_array(A) #Ensure A is non-empty 2D array containing only finite values.
    num_vecs = A.shape[1]
    effective_rank = num_vecs
    dtype = A.dtype
    Q = np.copy(A)
    R = np.zeros((num_vecs,num_vecs), dtype=dtype)

    _roundoff = 1e-8 #From reference paper
    _tau = 1e-2 #From reference paper

    if column_pivoting: #Initialize variables for fast pivoting, without re-evaluation of the norm at each step.
        squared_norms = weighted_norm(Q, M = M)**2
        max_norm = np.sqrt(np.max(squared_norms))
        columns_permutation = np.arange(num_vecs)

    for k in range(num_vecs):
        if column_pivoting:
            Q, R, squared_norms, columns_permutation = _column_pivot(Q, R, k, squared_norms, columns_permutation)
            norms_error_estimate = squared_norms * _roundoff
        if k != 0: #Reorthogonalization of the column k+1 of A with respect to the previous orthonormal k vectors.     
            alpha = weighted_dot_product(Q[:,:k], Q[:,k], M=M) #alpha = Q[:,:k].T@M@Q[:,k]
            R[:k,k] += alpha
            Q[:,k] -= np.dot(Q[:,:k],alpha)

        #Numerical rank detection, performed only when column_pivoting == True
        norm_at_iter_k = weighted_norm(Q[:,k], M=M)
        if column_pivoting:
            if norm_at_iter_k < rtol*max_norm:
                effective_rank = k
                if verbose:
                    warn("Numerical rank of A has been reached with a relative tolerance rtol = {:.2e}. Effective rank = {}. Stopping Orthogonalization procedure.".format(rtol, effective_rank))
                break    
        # Normalization of the column k + 1 
        R[k, k] = norm_at_iter_k
        Q[:, k] = Q[:, k] / R[k, k]
        # Orthogonalization of the remaining columns with respect to Q[:,k], i.e. the k+1 column of Q.
        if k < num_vecs - 1:
            R[k,k+1:] = weighted_dot_product(Q[:, k+1:], Q[:, k], M=M)       
            Q[:,k+1:] -= np.outer(Q[:, k], R[k,k+1:])
            if column_pivoting: #Try fast update of the squared norms, recompute if numerical criteria are not attained.
                squared_norms[k+1:] -= R[k,k+1:]**2 #Update norms using Phythagorean Theorem
                update_error_mask = _tau*squared_norms[k+1:] < norms_error_estimate[k+1:] #Check if the error estimate is too large
                if any(update_error_mask):
                    squared_norms[k+1:][update_error_mask] = weighted_norm(Q[:,k+1:][:,update_error_mask], M=M) #Recompute the norms if necessary.
    if column_pivoting:
        return Q[:,:effective_rank], R[:effective_rank], columns_permutation[:effective_rank]
    else:
        return Q[:,:effective_rank], R[:effective_rank]

class SquaredKernel(LinearOperator):
    r"""Helper class to repeatedly apply :math:`\alpha K^{2} + \beta K`.
    """    
    def __init__(self, K, alpha, beta):
        self.K = K
        self.is_linop = False
        if isinstance(K, LinearOperator):
            self.is_linop = True
        else:
            self.M = (alpha*(K@(K.T)) + beta*K) #Assuming that K is symmetric!
        self.dtype = K.dtype #Needed by LinearOperator superclass
        self.shape = K.shape #Needed by LinearOperator superclass
        self.alpha = alpha
        self.beta = beta

    def _matvec(self, x):
        if self.is_linop:
            v = np.ascontiguousarray(self.K @ x)
            return self.alpha * self.K @ v + self.beta * v
        else:
            return self.M@x

def randomized_range_finder(A_A_adj, size, n_iter, power_iteration_normalizer="none", M=None, random_state=None):
    """Randomized range finder. Adapted from sklearn.utils.extmath.randomized_range_finder

    Args:
        A_A_adj (ndarray, LinearOperator): Object representing :math:`AA^{*}`, where :math:`A` is the matrix whose range is to be found.
        size (int): Number of vectors to be generated and used to find the range of A.
        n_iter (int): Number of power iterations to be used to estimate the matrix range.
        power_iteration_normalizer (str, optional): Scheme to normalize the power iterations at each step. Defaults to "none". Available options are: "QR" and "column_pivoted_QR".
        M (ndarray or LinearOperator, optional): PSD linear operator. If not None, the vectors are orthonormalized with respect to the scalar product induced by M. Defaults to None corresponding to Identity matrix, i.e. standard Euclidean norm.
        random_state (int, optional):  int, RandomState instance or None, default=None. The seed of the pseudo random number generator to use when shuffling the data, i.e. getting the random vectors to initialize the algorithm. Pass an int for reproducible results across multiple function calls.

    Returns:
        ndarray: An (A_A_adj.shape[0], size) array the range of which approximates the range of A.
    """    
    #Adapted from sklearn.utils.extmath.randomized_range_finder
    A = aslinearoperator(A_A_adj)
    random_state = check_random_state(random_state)
    # Generating normal random vectors with shape: (A_A_adj.shape[0], size)
    assert size > 0, "Size must be greater than zero"
    Q = random_state.normal(size=(A_A_adj.shape[0], size))
    if hasattr(A_A_adj, "dtype") and A_A_adj.dtype.kind == "f":
        # Ensure f32 is preserved as f32
        Q = Q.astype(A_A_adj.dtype, copy=False)
    # Perform power iterations with Q to further 'imprint' the top
    # singular vectors of A in Q
    for _ in range(n_iter):
        if power_iteration_normalizer == "none":
            Q = A.matmat(Q)
        elif power_iteration_normalizer == "QR":
            Q, _ = modified_QR(A.matmat(Q), M=M)
        elif power_iteration_normalizer == "column_pivoted_QR":
            Q, _, _ = modified_QR(A.matmat(Q), M=M, column_pivoting=True)
    # Sample the range of A using by linear projection of Q
    # Extract an orthonormal basis
    Q, _, _ = modified_QR(A.matmat(Q), M=M, column_pivoting=True)
    return Q


from abc import ABCMeta, abstractmethod
import sklearn.gaussian_process.kernels as sk_kernels
from sklearn.metrics.pairwise import polynomial_kernel as sk_poly
from scipy.sparse.linalg import aslinearoperator
from ._keops_utils import Pm, lazy_cdist, lazy_cprod, __has_keops__, keops_import_error, __has_torch__
import numpy as np

 
from math import sqrt

def parse_backend(backend):
    if backend == 'keops':
        if __has_keops__:
            return backend
        else:
            raise keops_import_error
    elif backend == 'numpy':
        return backend
    else:
        raise ValueError('Invalid backend. Allowed values are \'numpy\' and \'keops\'.')

class Kernel(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, X, Y=None, backend='numpy'):
        """
        Evaluate the kernel. 
        The method called with backend == ``numpy'' returns a numpy array of shape (X.shape[0], Y.shape[0]).
        The method called with backend == ``keops'' returns an instance of a scipy linearoperator of shape (X.shape[0], Y.shape[0])
        """
    def __repr__(self):
        _r = "[" + self.__class__.__name__ + "] " 
        for k, v in self.__dict__.items():
            if k[0] == "_":
                pass
            else:
                _r += f"{k}: {v} "
        return _r

class ScalarProduct(Kernel):
    def __call__(self, X, Y = None, backend = 'numpy'):
        """Return the Kernel matrix k(x,y) := <Phi(x),Phi(y)>. Different samples are batched on the first dimension of X and Y."""
        phi_X = self.__feature_map__(X)

        if Y is None:
            phi_Y = phi_X
        else:
            phi_Y = self.__feature_map__(Y)
        if backend == 'numpy':
            return phi_X@(phi_Y.T)
        return lazy_cprod(phi_X, phi_Y) #Backend == 'keops'
    def cov(self, X, Y = None):
        phi_X = self.__feature_map__(X)
        if Y is None:
            return phi_X.T@phi_X
        else:
            if X.shape[0] != Y.shape[0]:
                raise ValueError("Shape mismatch: cross-covariances can be computed only if X.shape[0] == Y.shape[0] ")
            phi_Y = self.__feature_map__(Y)
            return phi_X.T@phi_Y
    @abstractmethod
    def __feature_map__(self, X):
        """Evaluate the feature map. The output should be a numpy array in _any_ case.""" 
        return X #Linear Kernel 

if __has_torch__:
    import torch
    from torch.nn import Module
    class TorchScalarProduct(Module):
        def __init__(self, feature_map):
            super().__init__()
            self.__feature_map__ = feature_map     
        def forward(self, X, Y=None, backend='numpy'):
            """
                Calculations are performed by pytorch, but if backend == 'numpy', numpy arrays are returned for consistency. 
                If backend == 'keops', an instance of a scipy linearoperator is returned instead.
            """
            with torch.no_grad():
                phi_X = self.__feature_map__(X)
                if Y is None:
                    phi_Y = phi_X
                else:
                    phi_Y = self.__feature_map__(Y)
                if backend == 'numpy':
                    return self.__to_numpy__(phi_X@(phi_Y.T))
                else:
                    return lazy_cprod(phi_X, phi_Y, backend='torch') #Backend == 'keops'
        def cov(self, X, Y = None):
            with torch.no_grad():
                phi_X = self.__feature_map__(X)
                if Y is None:
                    return self.__to_numpy__(phi_X.T@phi_X)
                else:
                    if X.shape[0] != Y.shape[0]:
                        raise ValueError("Shape mismatch: cross-covariances can be computed only if X.shape[0] == Y.shape[0] ")
                    phi_Y = self.__feature_map__(Y)
                    return self.__to_numpy__(phi_X.T@phi_Y)
        def __to_numpy__(self, tensor):
            return tensor.cpu().detach().numpy()

class RBF(Kernel):
    def __init__(self, length_scale=1.0):
        self.length_scale = length_scale 
        self._scikit_kernel =  sk_kernels.RBF(self.length_scale)  
        
    def __call__(self, X, Y=None, backend='numpy'):
        backend = parse_backend(backend)
        if backend == 'keops':   
            K = (-(lazy_cdist(X,Y)** 2) / (2*(Pm(self.length_scale)**2))).exp()
            return aslinearoperator(K)
        else:
            return self._scikit_kernel(X, Y)

class ExpSineSquared(Kernel):
    def __init__(self, length_scale=1.0, periodicity=1.0):
        self.length_scale = length_scale
        self.periodicity = periodicity
        self._scikit_kernel = sk_kernels.ExpSineSquared(length_scale=self.length_scale, periodicity=self.periodicity, length_scale_bounds='fixed', periodicity_bounds='fixed')
    def __call__(self, X, Y=None, backend='numpy'):
        backend = parse_backend(backend)
        if backend == 'keops':
            periodic = ((np.pi*lazy_cdist(X,Y)/Pm(self.periodicity)).sin()).square()
            K = (-(2*periodic)/((Pm(self.length_scale)**2))).exp()
            return aslinearoperator(K)
        else:
            return self._scikit_kernel(X, Y)

class Matern(Kernel):
    def __init__(self, nu=1.5, length_scale=1.0):
        self.nu = nu
        self.length_scale = length_scale
        self._scikit_kernel =  sk_kernels.Matern(self.length_scale, nu=self.nu)  

    def __call__(self, X, Y=None, backend='numpy'):
        backend = parse_backend(backend)  
        if backend == 'keops':  
            D = lazy_cdist(X,Y)/Pm(self.length_scale)
            if abs(self.nu - 0.5) <= 1e-12:
                K = (-D).exp()
            elif abs(self.nu - 1.5) <= 1e-12: #Once differentiable functions
                D *= sqrt(3)
                K = (1 + D)*((-D).exp())
            elif abs(self.nu - 2.5) <= 1e-12: #Twice differentiable functions
                D *= sqrt(5)
                K = (1 + D + (D**2)/3)*((-D).exp())
            else:
                raise(ValueError(f"Supported nu parameters are 0.5, 1.5, 2.5, while self.nu={self.nu}"))
            return aslinearoperator(K)
        else:
            return self._scikit_kernel(X, Y)
class Poly(Kernel):
    """
        Polynomial Kernel
        K(X, Y) = (gamma <X, Y> + coef0)^degree
    """
    def __init__(self, degree=3, gamma=None, coef0=1):
        self.gamma = gamma
        self.coef0 = coef0
        self.degree = degree

    def __call__(self, X, Y=None, backend='numpy'):
        backend = parse_backend(backend)  
        if self.gamma is None:
            _gamma = 1.0 / X.shape[1]
        else:
            _gamma = self.gamma
        
        if backend == 'keops':
            inner = Pm(self.coef0) + lazy_cprod(X,Y)*Pm(_gamma)
            if self.degree == 1:
                return aslinearoperator(inner)
            elif self.degree == 2:
                return aslinearoperator(inner.square())
            else:
                raise NotImplementedError("Poly kernel with degree != [1,2] not implemented (because of a bug on pow function).")
        else:
            return sk_poly(X, Y, degree = self.degree, gamma = self.gamma, coef0 = self.coef0)  
                     
class Linear(Poly):
    def __init__(self, gamma=None, coef0=1):
        super().__init__(degree=1, gamma=gamma, coef0=coef0)

class Quadratic(Poly):
    def __init__(self, gamma=None, coef0=1):
        super().__init__(degree=2, gamma=gamma, coef0=coef0)